/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CAOSDB_VALUE_H
#define CAOSDB_VALUE_H
#include <string>                           // for string
#include "caosdb/protobuf_helper.h"         // for ProtoMessageWrapper
#include "caosdb/entity/v1alpha1/main.pb.h" // for RepeatedPtrField, Message
namespace caosdb::entity {
using caosdb::utility::ProtoMessageWrapper;
using ProtoSpecialValue = caosdb::entity::v1alpha1::SpecialValue;
using ProtoValue = caosdb::entity::v1alpha1::Value;
using ValueCase = caosdb::entity::v1alpha1::Value::ValueCase;
using ScalarValueCase = caosdb::entity::v1alpha1::ScalarValue::ScalarValueCase;

class Entity;
class Property;

// Represents special values which are otherwise hard to tranfer via protobuf.
enum SpecialValue {
  // Represent the NULL value.
  NULL_VALUE = ProtoSpecialValue::SPECIAL_VALUE_UNSPECIFIED,
  // The empty string.
  EMPTY_STRING = ProtoSpecialValue::SPECIAL_VALUE_EMPTY_STRING,
};

class Value : public ProtoMessageWrapper<ProtoValue> {
public:
  inline Value() : ProtoMessageWrapper<ProtoValue>() {
    // has NULL_VALUE now
  }
  explicit inline Value(const std::string &value)
    : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_string_value(value);
  }
  explicit inline Value(double value) : ProtoMessageWrapper<ProtoValue>() {
    this->wrapped->mutable_scalar_value()->set_double_value(value);
  }
  explicit inline Value(const char value[]) : Value(std::string(value)) {}

  [[nodiscard]] inline auto IsString() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kStringValue) ||
             (this->wrapped->scalar_value().scalar_value_case() ==
                ScalarValueCase::kSpecialValue &&
              this->wrapped->scalar_value().special_value() ==
                ProtoSpecialValue::SPECIAL_VALUE_EMPTY_STRING);
    }
    return false;
  }
  [[nodiscard]] inline auto AsString() const noexcept -> const std::string & {
    return this->wrapped->scalar_value().string_value();
    ;
  }

  [[nodiscard]] inline auto IsDouble() const noexcept -> bool {
    if (this->wrapped->value_case() == ValueCase::kScalarValue) {

      return (this->wrapped->scalar_value().scalar_value_case() ==
              ScalarValueCase::kDoubleValue);
    }
    return false;
  }
  [[nodiscard]] inline auto AsDouble() const noexcept -> double {
    return this->wrapped->scalar_value().double_value();
  }

  inline auto operator==(const Value &other) const noexcept -> bool {
    return this->wrapped->SerializeAsString() ==
           other.wrapped->SerializeAsString();
  }

  friend class Entity;
  friend class Property;
};

} // namespace caosdb::entity

#endif
